import copy
from bs4 import BeautifulSoup


def table_span_normalize(html):
    if isinstance(html, str):
        soup = BeautifulSoup(html, 'html.parser')
        if soup is None:
            raise ValueError('Can\'t build DOM tree')
    elif not isinstance(html, BeautifulSoup):
        raise ValueError(f'Unsupported input type {type(html)}')
    else:
        soup = html

    for table in reversed(soup('table')):
        rows = []
        for thead in table('thead', recursive=False):
            rows.extend(thead('tr', recursive=False))
        for tbody_tr in table(['tbody', 'tr'], recursive=False):
            if 'tr' == tbody_tr.name:
                rows.append(tbody_tr)
            else:
                rows.extend(tbody_tr('tr', recursive=False))
        for tfoot in table('tfoot', recursive=False):
            rows.extend(tfoot('tr', recursive=False))

        rows_ = [[] for _ in range(len(rows))]
        for r, row in enumerate(rows):
            for cell in row(['th', 'td'], recursive=False):
                rowspan = int(cell.get('rowspan', 1))
                rowspan = len(rows) if 0 == rowspan else rowspan
                rowspan = min(rowspan, len(rows) - r)
                del cell['rowspan']

                colspan = int(cell.get('colspan', 1))
                colspan = 1 if 0 == r and len(rows) == rowspan else colspan
                colspan = max(colspan, 1)
                del cell['colspan']

                for rs in range(r, len(rows)):
                    rows_[rs].extend([None] * colspan)

                ix = rows_[r].index(None)

                for rs in range(r, r + rowspan):
                    for cs in range(ix, ix + colspan):
                        if rows_[rs][cs] is not None:
                            colspan = min(colspan, cs - ix)

                for rs in range(r, r + rowspan):
                    for cs in range(ix, ix + colspan):
                        cell_ = copy.deepcopy(cell)
                        rows_[rs][cs] = cell_

        maxcol = 1
        for row in rows_:
            for i, cell in enumerate(row):
                if cell is None:
                    continue
                maxcol = max(maxcol, i + 1)
        for r, row_ in enumerate(rows_):
            row_ = row_[:maxcol] + [None] * (maxcol - len(row_))
            row_ = [soup.new_tag('td') if c is None else c for c in row_]
            rows_[r] = row_

        drop = []
        for i in range(maxcol - 1):
            same = []
            for row_ in rows_:
                same.append(str(row_[i]) == str(row_[i + 1]))
            if same and all(same):
                drop.append(i + 1)
        for row_ in rows_:
            for i in drop[::-1]:
                del row_[i]


        for row, row_ in zip(rows, rows_):
            row.clear()
            for cell_ in row_:
                row.append(cell_)

    if isinstance(html, str):
        html = str(soup).strip()
    else:
        html = soup

    return html



import re
from rbloom import Bloom
from unicodedata import normalize

def make_dedup_bloom(separator='\n', just_words=True, zero_digits=True, capacity=100000000, error=0.00001):
    bf = Bloom(capacity, error)

    def _dedup_fn(text):
        if not isinstance(text, str):
            raise TypeError(f'Expected "text" to be a sting, found: {type(text)}')

        text = normalize('NFKD', text)

        lines = []
        for line in text.split(separator):
            line = line.strip()

            if not line:
                lines.append(line)
                continue

            if just_words:
                line = ' '.join(re.findall(r'\w+', line))
            if zero_digits:
                line = re.sub(r'\d', '0', line)

            if line in bf:
                line = ''
            else:
                bf.add(line)

            lines.append(line)

        return separator.join(lines)

    return _dedup_fn


import re
from unicodedata import normalize

ACCENT_GLYPHS = re.compile('[\u0060\u02C6\u02CA\u02CB\u02CE\u02CF\u02DD\u02DF\u02F4\u02F5\u02F6\u0300\u0301\u0302]+')


def remove_accent_glyphs(text):
    text = normalize('NFD', text)
    text = re.sub(ACCENT_GLYPHS, '', text)
    text = normalize('NFC', text)

    return text



import markitdown
import re
import tempfile
import trafilatura
from bs4 import BeautifulSoup, Comment
from markdownify import MarkdownConverter
from nlpclean.table import table_span_normalize

_MEANINGLESS_TAGS = [
    'applet', 'audio', 'canvas', 'comment', 'datalist', 'embed', 'form',
    'frame', 'frameset', 'iframe', 'img', 'kbd', 'map', 'menu', 'noembed',
    'noframes', 'noscript', 'object', 'output', 'plaintext', 'script', 'style',
    'svg', 'template', 'title', 'video', 'xmp'
]

_BLOCKLEVEL_TAGS = [
    'address', 'article', 'aside', 'blockquote', 'center', 'dd', 'details',
    'dialog', 'dir', 'div', 'dl', 'dt', 'fieldset', 'figcaption', 'figure',
    'footer', 'form', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'header', 'hgroup',
    'hr', 'isindex', 'li', 'main', 'menu', 'nav', 'noframes', 'noscript', 'ol',
    'p', 'pre', 'section', 'table', 'td', 'th', 'tr', 'ul'
]


def html_to_article(html, parser='lxml'):
    soup = BeautifulSoup(html, parser)
    if soup is None:
        raise ValueError('Can\'t build DOM tree')

    # Drop comments
    for comment in soup(string=lambda txt: isinstance(txt, Comment)):
        comment.extract()

    # Drop non-meaning tags
    for node in soup(_MEANINGLESS_TAGS):
        node.extract()

    # Drop <pre> wrapper around <code>
    for code in soup('code'):
        if 'pre' != code.parent.name:
            continue
        if 1 != len(code.parent(recursive=False)):
            continue
        code.parent.replace_with(code)

    # Normalize tables
    soup = table_span_normalize(soup)

    html = str(soup)

    extract_kwargs = {
        'filecontent': html,
        'favor_recall': True,
        'output_format': 'html',
    }
    html = trafilatura.extract(**extract_kwargs)

    soup = BeautifulSoup(html, parser)
    if soup is None:
        raise ValueError('Can\'t build DOM tree')

    # Revert <tr> and <td>
    for node in soup('row'):
        node.name = 'tr'
    for node in soup('cell'):
        node.name = 'td'

    html = str(soup)

    html = html.replace('\u0097', '').replace('\u200c', '')

    return html


def fragment_to_markdown(html, parser='lxml'):
    if isinstance(html, str):
        if not len(html.strip()):
            return ''
        soup = BeautifulSoup(html, parser)
        if soup is None:
            raise ValueError('Can\'t build DOM tree')
    elif not isinstance(html, BeautifulSoup):
        raise ValueError(f'Unsupported input type {type(html)}')
    else:
        soup = html

    soup = table_span_normalize(soup)

    converter = MarkdownConverter(strip=['a', 'img'], heading_style='atx')
    md = converter.convert_soup(soup).strip()

    md = re.sub(
        '```.+?```',
        lambda m: m.group(0).strip().replace('\n', '___NEWLINE_HACK___'),
        md, flags=re.S)
    md = re.sub(r'[^\S\n]*\n[^\S\n]*', '\n', md)
    md = re.sub(r'\n{3,}', '\n\n', md)
    md = md.replace('___NEWLINE_HACK___', '\n')
    md = re.sub(r'\n```\n*(.+?)\n*```\n', r'\n```\n\1\n```\n', md, flags=re.S)

    md = md.replace('\n\n— ', '\n* ')
    md = md.replace('\n— ', '\n* ')

    return md


def fragment_to_text(html, parser='lxml'):
    if isinstance(html, str):
        if not len(html.strip()):
            return ''
        soup = BeautifulSoup(html, parser)
        if soup is None:
            raise ValueError('Can\'t build DOM tree')
    elif not isinstance(html, BeautifulSoup):
        raise ValueError(f'Unsupported input type {type(html)}')
    else:
        soup = html

    # Drop comments
    for comment in soup(string=lambda txt: isinstance(txt, Comment)):
        comment.extract()

    # Drop non-meaning tags
    for node in soup(_MEANINGLESS_TAGS):
        node.replace_with(soup.new_tag('br'))

    # Insert linebreaks around block-level tags
    for node in soup(_BLOCKLEVEL_TAGS):
        node.insert_before(soup.new_tag('br'))
        node.insert_before(soup.new_tag('br'))
        node.insert_after(soup.new_tag('br'))
        node.insert_after(soup.new_tag('br'))

    # Remove linebreaks inside text nodes (as browser does)
    for node in soup(string=lambda string: '\n' in string):
        node.string.replace_with(node.string.replace('\n', ' '))

    # Swap html linebreaks to normal ones
    for node in soup('br'):
        node.replace_with('\n')

    # Normalize tables
    soup = table_span_normalize(soup)

    # Cleanup final text
    text = soup.getText()
    text = str(text).strip()
    text = re.sub(' {2,}', ' ', text)
    text = text.replace(' \n', '\n').replace('\n ', '\n')
    text = re.sub('\n{3,}', '\n\n', text)

    return text



import fasttext
import iso639
import logging
import os
import requests
import tempfile
import tqdm


def _download_fasttext():
    remote_url = 'https://huggingface.co/cis-lmu/glotlid/resolve/main/model.bin'
    expected_size = 1687094687

    model_dir = os.path.expanduser(os.path.join('~', '.nlpclean'))
    model_dir = os.getenv("DATA_DIR") or model_dir
    os.makedirs(model_dir, exist_ok=True)

    model_name = 'model.bin'
    model_path = os.path.join(model_dir, model_name)

    if os.path.exists(model_path) and os.path.getsize(model_path) == expected_size:
        return model_path

    remote_size = int(
        requests.head(remote_url, allow_redirects=True).headers['Content-Length'])
    if expected_size != remote_size:
        raise AssertionError('Unexpected remote model size')

    with tqdm.tqdm(initial=0, total=remote_size, unit='B', unit_scale=True, desc=model_name) as bar:
        with tempfile.NamedTemporaryFile(delete=False) as f:
            temp_name = f.name

            request = requests.get(remote_url, allow_redirects=True, stream=True)
            for chunk in request.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
                    bar.update(1024)

    os.rename(temp_name, model_path)

    return model_path


def detect_main_lang():
    _fasttext_model = fasttext.load_model(_download_fasttext())

    def _detect_fn(text):
        if not text.strip():
            return '??', 1.0

        try:
            scores = _fasttext_model.predict(text.replace('\n', ' '))
        except Exception as e:
            logging.warning(e)
            return '??', 0.0

        if not scores or not scores[0]:
            return '??', 0.0

        lang = scores[0][0].replace('__label__', '').split('_')[0]
        lang = iso639.Lang(lang).pt1
        if not lang:
            return '??', scores[1][0]

        return lang, scores[1][0]

    return _detect_fn
